// Copyright 2017 SiegeLord
// Licensed under GPL, see LICENSE for full terms

#![allow(dead_code)]

use allegro::*;
use std::rc::Rc;
use game_state::GameState;
use util::load_config;
use std::path::Path;
use errors::Res;

slr_def!
{
	pub struct AnimationInfo
	{
		pub width: f32 = 0.0,
		pub height: f32 = 0.0,
		pub rate: f64 = 0.0,
		pub file: String = "".to_string(),
	}
}

pub struct Animation
{
	bmp: Rc<Bitmap>,
	offset_time: f64,
	play_once: bool,
	num_frames: i32,
	num_frames_x: i32,
	info: AnimationInfo,
}

impl Animation
{
	pub fn new(filename: &str, play_once: bool, state: &mut GameState) -> Res<Animation>
	{
		let path = Path::new(filename);
		Ok(if path.extension().unwrap().to_str().unwrap() == "png"
		{
			let bmp = state.bitmap_cache.load(&state.core, filename)?;
			let width = bmp.get_width() as f32;
			let height = bmp.get_height() as f32;
			Animation
			{
				bmp: bmp,
				offset_time: state.core.get_time(),
				play_once: play_once,
				num_frames: 1,
				num_frames_x: 1,
				info: AnimationInfo
				{
					width: width,
					height: height,
					rate: 0.0,
					file: filename.to_string(),
				}
			}
		}
		else
		{
			let mut info = AnimationInfo::new();
			load_config(filename, &mut info).unwrap();
			
			let bmp = state.bitmap_cache.load(&state.core, &info.file)?;
			
			let num_frames_x = (bmp.get_width() as f32 / info.width) as i32;
			let num_frames_y = (bmp.get_height() as f32 / info.height) as i32;
			
			Animation
			{
				bmp: bmp,
				offset_time: state.core.get_time(),
				play_once: play_once,
				num_frames: num_frames_x * num_frames_y,
				num_frames_x: num_frames_x,
				info: info,
			}
		})
	}
	
	pub fn reset_time(&mut self, core: &Core)
	{
		self.offset_time = core.get_time();
	}

	pub fn draw(&self, x: f32, y: f32, core: &Core)
	{
		let raw_frame = (self.num_frames as f64 * (core.get_time() - self.offset_time) * self.info.rate) as i32;
		if self.play_once && raw_frame >= self.num_frames
		{
			return;
		}
		let frame = raw_frame % self.num_frames;
		let sx = (frame % self.num_frames_x) as f32 * self.info.width;
		let sy = (frame / self.num_frames_x) as f32 * self.info.height;
		core.draw_bitmap_region(&*self.bmp, sx, sy, self.info.width, self.info.height, x, y, Flag::zero());
	}

	pub fn get_width(&self) -> f32
	{
		self.info.width
	}

	pub fn get_height(&self) -> f32
	{
		self.info.height
	}
}
