// Copyright 2017 SiegeLord
// Licensed under GPL, see LICENSE for full terms

use allegro::*;
use allegro_font::*;

use engine::id_map::HasId;
use std::collections::{HashMap, BinaryHeap};
use game_state::{Vec2D, Object, GameState, Collision, CollisionClass, Team};
use tiled::{parse_file, Map, ObjectShape};
use std::path::Path;
use errors::*;
use util::load_config;
use std::rc::{Rc, Weak};
use actor::create_succubus;
use ai::{create_soldier, create_devil};
use std::cmp::{min, max};
use util::l1_dist;
use slr_config::ElementRepr;

use std::fs::File;
use std::fs::read_dir;
use std::io::Write;

slr_def!
{
	#[derive(Clone)]
	pub struct Level
	{
		pub map: String = "".into(),
		pub high_score: i32 = 0,
		pub name: String = "".into()
	}
}

struct Tile
{
	bitmap: Weak<SubBitmap>,
}

#[derive(Copy, Clone, Eq, PartialEq)]
enum State
{
	Playing,
	Won,
	Lost
}

pub struct GameMode
{
	level_filename: String,
	level: Level,
	map: Map,
	tiles: HashMap<u32, Tile>,
	collision_grid: Vec<usize>,
	wall_collision_grid: Vec<usize>,
	num_devils: i32,
	num_soldiers: i32,
	heart_icon: Rc<Bitmap>,
	soldier_icon: Rc<Bitmap>,
	devil_icon: Rc<Bitmap>,
	state: State,
	pub score: i32,
}

fn get_tile_pos(pos: Vec2D) -> (i32, i32)
{
	(((pos.x + 16.0) / 32.0).floor() as i32,
	 ((pos.y + 16.0) / 32.0).floor() as i32)
}

impl GameMode
{
	pub fn get_local_move_goal(&self, id: usize, goal_id: usize, cur_pos: Vec2D, goal_pos: Vec2D) -> Vec2D
	{
		let (cur_x, cur_y) = get_tile_pos(cur_pos);
		let (goal_x, goal_y) = get_tile_pos(goal_pos);
		
		#[derive(Ord, PartialOrd, PartialEq, Eq)]
		struct Node
		{
			// Since this is a max heap, we'll store negative scores.
			neg_cost: i32,
			x: i32,
			y: i32,
		}
		#[derive(Ord, PartialOrd, PartialEq, Eq)]
		struct ApproxNode
		{
			// Since this is a max heap, we'll store negative scores.
			neg_approx_cost: i32,
			neg_true_cost: i32,
			x: i32,
			y: i32,
		}
		
		let mut open_set = BinaryHeap::new();
		let mut cost_map = HashMap::new();
		let mut approx_cost_heap = BinaryHeap::new();
		let mut came_from = HashMap::new();
		open_set.push(Node
		{
			neg_cost: 0,
			x: cur_x,
			y: cur_y,
		});
		approx_cost_heap.push(ApproxNode
		{
			neg_approx_cost: -l1_dist(cur_x, cur_y, goal_x, goal_y),
			neg_true_cost: 0,
			x: cur_x,
			y: cur_y,
		});
		cost_map.insert((cur_x, cur_y), 0);
		
		let mut found = None;
		while !open_set.is_empty()
		{
			let best = open_set.pop().unwrap();
			let cost = *cost_map.get(&(best.x, best.y)).unwrap();
			if best.x == goal_x && best.y == goal_y
			{
				found = Some((best.x, best.y));
				break;
			}
			for m in &[(-1, 0), (1, 0), (0, -1), (0, 1)]
			{
				let cand_x = best.x + m.0;
				let cand_y = best.y + m.1;
				let cand_cost = cost + 1;
				//~ info!("trying: {} {} at {}", cand_x, cand_y, cand_cost);
				if cost_map.get(&(cand_x, cand_y)).map_or(true, |&v| cand_cost < v) &&
					(!self.solid(id, cand_x, cand_y) || !self.solid(goal_id, cand_x, cand_y))
					//((cand_x == cur_x && cand_y == cur_y) || !self.solid(cand_x, cand_y))
				{
					let approx_cost = l1_dist(goal_x, goal_y, cand_x, cand_y);
					open_set.push(Node
					{
						neg_cost: -(cand_cost + approx_cost),
						x: cand_x,
						y: cand_y,
					});
					approx_cost_heap.push(ApproxNode
					{
						neg_approx_cost: -approx_cost,
						neg_true_cost: -cand_cost,
						x: cand_x,
						y: cand_y,
					});
					cost_map.insert((cand_x, cand_y), cand_cost);
					came_from.insert((cand_x, cand_y), (best.x, best.y));
				}
			}
		}

		let mut cur;
		let mut prev;
		match found
		{
			Some(found) => cur = found,
			None =>
			{
				if let Some(approx_best) = approx_cost_heap.pop()
				{
					cur = (approx_best.x, approx_best.y);
					//~ info!("Got {:?}", cur);
				}
				else
				{
					//~ info!("Got nothing");
					return cur_pos;
				}
			}
		}
		prev = cur;
		loop
		{
			if cur.0 == cur_x && cur.1 == cur_y
			{
				return Vec2D::new(prev.0 as f32 * 32.0, prev.1 as f32 * 32.0);
			}
			prev = cur;
			cur = *came_from.get(&cur).unwrap();
		}
	}
	
	pub fn clear_shot(&self, cur_pos: Vec2D, goal_pos: Vec2D, max_range: i32) -> bool
	{
		let (cur_x, cur_y) = get_tile_pos(cur_pos);
		let (goal_x, goal_y) = get_tile_pos(goal_pos);
		
		if cur_x == goal_x
		{
			let start_y = min(cur_y, goal_y);
			let end_y = max(cur_y, goal_y);
			if end_y - start_y > max_range
			{
				return false;
			}
			for y in start_y..end_y
			{
				if self.wall_solid(cur_x, y)
				{
					return false;
				}
			}
			true
		}
		else if cur_y == goal_y
		{
			let start_x = min(cur_x, goal_x);
			let end_x = max(cur_x, goal_x);
			if end_x - start_x > max_range
			{
				return false;
			}
			for x in start_x..end_x
			{
				if self.wall_solid(x, cur_y)
				{
					return false;
				}
			}
			true
		}
		else
		{
			false
		}
	}
	
	pub fn solid(&self, id: usize, x: i32, y: i32) -> bool
	{
		let x = min(max(x, 0), self.map.width as i32 - 1);
		let y = min(max(y, 0), self.map.height as i32 - 1);
		
		let val = self.collision_grid[(x + y * self.map.width as i32) as usize];
		
		return val != 0 && val != id;
	}
	
	pub fn wall_solid(&self, x: i32, y: i32) -> bool
	{
		let x = min(max(x, 0), self.map.width as i32 - 1);
		let y = min(max(y, 0), self.map.height as i32 - 1);
		
		let val = self.wall_collision_grid[(x + y * self.map.width as i32) as usize];
		
		return val != 0;
	}
}

pub fn create_wall(parent: usize, pos: Vec2D, width: f32, height: f32, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.collision = Some(Collision{ tl: Vec2D::new(0.0, 0.0), br: Vec2D::new(width, height) });
	obj.collision_class = CollisionClass::Wall;
	
	Ok(obj)
}

pub fn compute_collision_grid(map: &Map, collisions: &[(usize, Collision)]) -> Vec<usize>
{
	let mut collision_grid = vec![];
	for y in 0..map.height
	{
		for x in 0..map.width
		{
			let test_collision = Collision{ tl: Vec2D::new(8.0, 8.0), br: Vec2D::new(32.0 - 4.0, 32.0 - 4.0) };
			let pos = Vec2D::new((x * map.tile_width) as f32, (y * map.tile_height) as f32);
			let test_collision = test_collision.add_pos(pos);
			let mut collides = 0;
			for &(id, collision) in collisions
			{
				if collision.collides_with(&test_collision)
				{
					collides = id;
					break;
				}
			}
			collision_grid.push(collides);
		}
	}
	collision_grid
}

pub fn create_game_mode(level_file: &str, state: &mut GameState) -> Res<Object>
{
	let game_mode_id = state.id_mint.new_id();
	
	let mut level = Level::new();
	
	load_config(level_file, &mut level)?;
	
	let map_path = Path::new(&level.map).to_owned();
	
	let map = parse_file(&map_path)?;
	
	// TODO: This offset should be read from the map.
	let offset = Vec2D::new(-16.0, 16.0);
	
	let (tiles, collision_grid) = {
		
		let mut wall_collisions = vec![];
		
		for obj_group in &map.object_groups
		{
			for obj_def in &obj_group.objects
			{
				let obj = match obj_def.name.as_str()
				{
					"start" =>
					{
						create_succubus(game_mode_id.get(), Vec2D::new(obj_def.x + offset.x, obj_def.y + offset.y), state)
					},
					"soldier" =>
					{
						create_soldier(game_mode_id.get(), Vec2D::new(obj_def.x + offset.x, obj_def.y + offset.y), state)
					},
					"devil" =>
					{
						create_devil(game_mode_id.get(), Vec2D::new(obj_def.x + offset.x, obj_def.y + offset.y), state)
					},
					"wall" =>
					{
						let (width, height) = match obj_def.shape
						{
							ObjectShape::Rect{width, height} => (width, height),
							_ => (0.0, 0.0)
						};
						
						let pos = Vec2D::new(obj_def.x + offset.x, obj_def.y + offset.y);
						let collision = Collision{ tl: Vec2D::new(0.0, 0.0), br: Vec2D::new(width, height) };
						let collision = collision.add_pos(pos);

						let wall = create_wall(game_mode_id.get(), pos, width, height, state)?;
						
						wall_collisions.push((wall.get_id(), collision));
						
						Ok(wall)
					},
					_ => continue
				}?;
				
				state.add_object(obj);
			}
		}
		
		let collision_grid = compute_collision_grid(&map, &wall_collisions);
		
		let parent_path = map_path.parent().unwrap_or(Path::new(""));
		let tileset = &map.tilesets[0];
		let image = &tileset.images[0];
		let bitmap_path = parent_path.join(Path::new(&image.source));
		let bmp = state.bitmap_cache.load(&state.core, bitmap_path.to_str().unwrap())?;
		
		let num_cols = image.width / tileset.tile_width as i32;
		
		let mut tiles = HashMap::new();
	
		for tile in &tileset.tiles
		{		
			let x_tile = tile.id % num_cols as u32;
			let y_tile = tile.id / num_cols as u32;
			
			let x = (x_tile * tileset.tile_width) as i32;
			let y = (y_tile * tileset.tile_height) as i32;
			
			let sub_bitmap = bmp.create_sub_bitmap(x, y, tileset.tile_width as i32, tileset.tile_height as i32).unwrap();
			
			tiles.insert(tile.id + tileset.first_gid, Tile{ bitmap: sub_bitmap });
		}
		
		(tiles, collision_grid)
	};
	
	state.sfx.play_music("data/musix-the-red-one.mod", &state.audio);
	
	let mut obj = Object::new(game_mode_id);
	
	obj.game_mode = Some(GameMode
	{
		level_filename: level_file.into(),
		level: level,
		map: map,
		tiles: tiles,
		collision_grid: collision_grid.clone(),
		wall_collision_grid: collision_grid,
		num_soldiers: 0,
		num_devils: 0,
		heart_icon: state.bitmap_cache.load(&state.core, "data/heart_icon.png")?,
		soldier_icon: state.bitmap_cache.load(&state.core, "data/soldier_icon.png")?,
		devil_icon: state.bitmap_cache.load(&state.core, "data/devil_icon.png")?,
		state: State::Playing,
		score: 0,
	});
	
	Ok(obj)
}

simple_behavior!
{
	GameModeDraw[obj.game_mode.is_some()] |obj, state|
	{
		let obj = obj.borrow();
		
		let game_mode = obj.game_mode.as_ref().unwrap();
		for layer in &game_mode.map.layers
		{
			for (y_idx, row) in layer.tiles.iter().enumerate()
			{
				for (x_idx, tile_id) in row.iter().enumerate()
				{
					let x = x_idx as f32 * game_mode.map.tile_width as f32;
					let y = y_idx as f32 * game_mode.map.tile_height as f32;
					
					if let Some(tile) = game_mode.tiles.get(tile_id)
					{
						state.core.draw_bitmap(&*tile.bitmap.upgrade().unwrap(), x, y, Flag::zero());
					}
				}
			}
		}
	}
}

simple_behavior!
{
	GameModeInput[obj.game_mode.is_some()] |obj, state|
	{
		let obj = obj.borrow();
		
		let game_mode = obj.game_mode.as_ref().unwrap();
		
		if let Some(key) = state.key_down
		{
			if key == KeyCode::R
			{
				state.remove_object(obj.get_id());
				let new_game = create_game_mode(&game_mode.level_filename, state)?;
				state.paused = false;
				state.start_fade = true;
				state.add_object(new_game);
			}
			else if key == KeyCode::P && game_mode.state == State::Playing
			{
				state.paused = !state.paused;
				if state.paused
				{
					state.sfx.play_music("data/musix-retrospective.ogg", &state.audio);
				}
				else
				{
					state.sfx.play_music("data/musix-the-red-one.mod", &state.audio);
				}
			}
			else if key == KeyCode::Escape
			{
				state.remove_object(obj.get_id());
				let mode = ::menu_mode::create_menu_mode(state)?;
				state.paused = false;
				state.start_fade = true;
				state.add_object(mode);
			}
		}
	}
}

pub struct GameModeUIDraw;

complex_behavior!
{
	GameModeUIDraw[true] |self, _obj, objs, state|
	{
		let mut game_mode_id = None;
		let mut num_hearts = 0;
		
		for obj in objs.elems()
		{
			let obj = obj.borrow();
			if obj.game_mode.is_some()
			{
				game_mode_id = Some(obj.get_id());
			}
			
			if obj.is_player
			{
				num_hearts = obj.num_hearts;
			}
		}
		
		if let Some(game_mode_id) = game_mode_id
		{
			let obj = objs.get(game_mode_id).unwrap();
			let obj = obj.borrow();
			let game_mode = obj.game_mode.as_ref().unwrap();
			let core = &state.core;
			
			match game_mode.state
			{
				State::Playing => 
				{
					core.draw_bitmap(&*game_mode.heart_icon, 14.0, 16.0, Flag::zero());
					core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						32.0, 16.0, FontAlign::Left, &format!("{}", num_hearts));
					
					core.draw_bitmap(&*game_mode.devil_icon, ::BUFFER_WIDTH as f32 / 2.0 - 18.0, 16.0, Flag::zero());
					core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						::BUFFER_WIDTH as f32 / 2.0, 16.0, FontAlign::Left, &format!("{}", game_mode.num_devils));
					
					core.draw_bitmap(&*game_mode.soldier_icon, ::BUFFER_WIDTH as f32 - 50.0, 16.0, Flag::zero());
					core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						::BUFFER_WIDTH as f32 - 32.0, 16.0, FontAlign::Left, &format!("{}", game_mode.num_soldiers));
					if state.paused == true
					{
						core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
									::BUFFER_WIDTH as f32 / 2.0, ::BUFFER_HEIGHT as f32 / 2.0, FontAlign::Centre,
									"Paused");
					}
				},
				State::Lost =>
				{
					state.prim.draw_filled_rectangle(0.0, 0.0, ::BUFFER_WIDTH as f32, ::BUFFER_HEIGHT as f32, Color::from_rgba_f(0.5, 0.0, 0.0, 0.5));
					core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						::BUFFER_WIDTH as f32 / 2.0, ::BUFFER_HEIGHT as f32 / 2.0, FontAlign::Centre,
						"My soldiers perished! Demon lords will be displeased!");
					core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						::BUFFER_WIDTH as f32 / 2.0, ::BUFFER_HEIGHT as f32 / 2.0 + 16.0, FontAlign::Centre,
						"Press R to restart.");
				},
				State::Won =>
				{
					state.prim.draw_filled_rectangle(0.0, 0.0, ::BUFFER_WIDTH as f32, ::BUFFER_HEIGHT as f32, Color::from_rgba_f(0.0, 0.0, 0.0, 0.5));
					core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						::BUFFER_WIDTH as f32 / 2.0, ::BUFFER_HEIGHT as f32 / 2.0, FontAlign::Centre,
						"Success! My pets have driven back the devilish attack!");
					core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						::BUFFER_WIDTH as f32 / 2.0, ::BUFFER_HEIGHT as f32 / 2.0 + 16.0, FontAlign::Centre,
						"Press Esc to go back to main menu.");
				}
			}
			
			core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						16.0, ::BUFFER_HEIGHT as f32 - 32.0, FontAlign::Left, &format!("Score: {}", game_mode.score));

			core.draw_text(&state.ui_font, Color::from_rgb(255, 255, 255),
						::BUFFER_WIDTH as f32 - 192.0, ::BUFFER_HEIGHT as f32 - 32.0, FontAlign::Left, &format!("High Score: {}", game_mode.level.high_score));
		}
	}
}

pub struct GameModeLogic;

complex_behavior!
{
	GameModeLogic[true] |self, _obj, objs, state|
	{
		let mut game_mode_id = None;
		let mut num_devils = 0;
		let mut num_soldiers = 0;
		let mut collisions = vec![];
		let mut wall_collisions = vec![];
		
		for obj in objs.elems()
		{
			let obj = obj.borrow();
			if obj.game_mode.is_some()
			{
				game_mode_id = Some(obj.get_id());
			}
			
			if obj.is_ai && obj.team.is_some()
			{
				match obj.team.unwrap()
				{
					Team::Soldier => num_soldiers += 1,
					Team::Devil => num_devils += 1,
				}
			}
			
			if obj.pos.is_some() && obj.collision.is_some() && obj.collision_class == CollisionClass::Wall || obj.collision_class == CollisionClass::Other
			{
				let collision = obj.collision.unwrap().add_pos(obj.pos.unwrap());
				collisions.push((obj.get_id(), collision));
				if obj.collision_class == CollisionClass::Wall
				{
					wall_collisions.push((obj.get_id(), collision));
				}
			}
		}
		
		if let Some(game_mode_id) = game_mode_id
		{
			let obj = objs.get(game_mode_id).unwrap();
			let mut obj = obj.borrow_mut();
			let game_mode = obj.game_mode.as_mut().unwrap();
			
			game_mode.collision_grid = compute_collision_grid(&game_mode.map, &collisions);
			game_mode.wall_collision_grid = compute_collision_grid(&game_mode.map, &wall_collisions);
			
			if num_devils == 0
			{
				game_mode.state = State::Won;
				if game_mode.num_devils != 0
				{
					state.paused = true;
					state.play_sound("data/victory.wav")?;
					
					if game_mode.score > game_mode.level.high_score
					{
						let mut level = game_mode.level.clone();
						level.high_score = game_mode.score;
						
						let mut f = File::create(&game_mode.level_filename).map_err(|e| format!("{}: {}", e.to_string(), game_mode.level_filename))?;
						
						let elem = level.to_element();
						
						write!(f, "{}", elem).map_err(|e| format!("{}: {}", e.to_string(), game_mode.level_filename))?;
					}
				}
			}
			else if num_soldiers == 0
			{
				game_mode.state = State::Lost;
				if game_mode.num_soldiers != 0
				{
					state.play_sound("data/defeat.wav")?;
					state.paused = true;
				}
			}
			game_mode.num_devils = num_devils;
			game_mode.num_soldiers = num_soldiers;
		}
	}
}

pub struct MapDef
{
	pub level: Level,
	pub level_filename: String,
}

pub fn load_maps(dir: &str) -> Res<Vec<MapDef>>
{
	let files = read_dir(&Path::new(dir)).unwrap();
	let mut ret: Vec<_> = files.filter(|p|
	{
		p.as_ref().unwrap().path().extension().unwrap().to_str().unwrap() == "cfg"
	}).map(|p|
	{
		let p = p.as_ref().unwrap().path();
		let filename = p.to_str().unwrap().to_string();
		let mut level = Level::new();
		load_config(&filename, &mut level).unwrap();
		MapDef
		{
			level: level,
			level_filename: filename
		}
	}).collect();
	ret.sort_by_key(|v| v.level_filename.clone());
	assert!(ret.len() > 0, format!("No levels found in '{}'!", dir));
	Ok(ret)
}
