// Copyright 2017 SiegeLord
// Licensed under GPL, see LICENSE for full terms

#![allow(dead_code)]

use allegro::KeyCode;
use allegro::Core;
use allegro_font::{FontAddon, Font};
use allegro_audio::AudioAddon;
use allegro_primitives::PrimitivesAddon;
use animation::Animation;
use engine::world::WorldState;
use engine::bitmap_cache::BitmapCache;
use engine::sample_cache::SampleCache;
use engine::id_map::{HasId, IdMint, UniqueId};
use cgmath::Vector2;
use std::cell::RefCell;
use errors::Res;
use game_mode::GameMode;
use menu_mode::MenuMode;
use intro_mode::IntroMode;
use sfx::Sfx;

pub type Vec2D = Vector2<f32>;

macro_rules! simple_behavior
{
	($name: ident[$check: expr] |$obj: ident, $state: ident| $e: expr) =>
	{
		pub struct $name;

		impl ::engine::world::Behavior<::std::cell::RefCell<::game_state::Object>, ::game_state::GameState> for $name
		{
			fn check_object(&self, $obj: &::std::cell::RefCell<::game_state::Object>) -> bool
			{
				let $obj = $obj.borrow();
				$check
			}

			fn handle_objects(&mut self, objects: &mut ::engine::id_map::IdMap<::std::cell::RefCell<::game_state::Object>>, $state: &mut ::game_state::GameState) -> ::errors::Res<()>
			{
				for $obj in objects.elems()
				{
					if self.check_object($obj)
					{
						$e
					}
				}
				Ok(())
			}
		}
	}
}

macro_rules! complex_behavior
{
	($name: ident[$check: expr] |$self_: ident, $obj: ident, $objects: ident, $state: ident| $e: expr) =>
	{
		impl ::engine::world::Behavior<::std::cell::RefCell<::game_state::Object>, ::game_state::GameState> for $name
		{
			fn check_object(&$self_, $obj: &::std::cell::RefCell<::game_state::Object>) -> bool
			{
				let $obj = $obj.borrow();
				$check
			}

			fn handle_objects(&mut $self_, $objects: &mut ::engine::id_map::IdMap<::std::cell::RefCell<::game_state::Object>>, $state: &mut ::game_state::GameState) -> ::errors::Res<()>
			{
				$e
				Ok(())
			}
		}
	}
}

macro_rules! object
{
	($name: ident
	{
		$($member: ident : $type_ : ty = $init: expr),* $(,)*
	}) =>
	{
		pub struct $name
		{
			id: UniqueId,
			$(pub $member : $type_),*
		}

		impl $name
		{
			pub fn new(id: UniqueId) -> $name
			{
				$name
				{
					id: id,
					$($member : $init),*
				}
			}
		}
	}
}

#[derive(Copy, Clone, Debug, Eq, PartialEq)]
pub enum Team
{
	Soldier,
	Devil,
}

#[derive(Copy, Clone, Debug)]
pub enum FaceDirection
{
	Up,
	Down,
	Left,
	Right
}

impl FaceDirection
{
	pub fn to_vec(&self) -> Vec2D
	{
		match *self
		{
			FaceDirection::Up => Vec2D::new(0.0, -1.0),
			FaceDirection::Down => Vec2D::new(0.0, 1.0),
			FaceDirection::Left => Vec2D::new(-1.0, 0.0),
			FaceDirection::Right => Vec2D::new(1.0, 0.0),
		}
	}
}

#[derive(Copy, Clone)]
pub struct Collision
{
	pub tl: Vec2D,
	pub br: Vec2D,
}

impl Collision
{
	pub fn collides_with(&self, other: &Collision) -> bool
	{
		return
			self.br.x > other.tl.x &&
			self.br.y > other.tl.y &&
			self.tl.x < other.br.x &&
			self.tl.y < other.br.y;
	}
	
	pub fn add_pos(&self, pos: Vec2D) -> Collision
	{
		Collision
		{
			tl: self.tl + pos,
			br: self.br + pos
		}
	}
}

pub struct Animations
{
	pub default: Option<Animation>,
	pub dancing: Option<Animation>,
	pub down: Option<Animation>,
	pub up: Option<Animation>,
	pub left: Option<Animation>,
	pub right: Option<Animation>,
	pub kiss_down: Option<Animation>,
	pub kiss_up: Option<Animation>,
	pub kiss_left: Option<Animation>,
	pub kiss_right: Option<Animation>,
	pub walk_down: Option<Animation>,
	pub walk_up: Option<Animation>,
	pub walk_left: Option<Animation>,
	pub walk_right: Option<Animation>,
}

impl Animations
{
	pub fn new() -> Self
	{
		Self
		{
			default: None,
			dancing: None,
			down: None,
			up: None,
			left: None,
			right: None,
			kiss_down: None,
			kiss_up: None,
			kiss_left: None,
			kiss_right: None,
			walk_down: None,
			walk_up: None,
			walk_left: None,
			walk_right: None,
		}
	}
}

pub fn succubus_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/succubus_down.png", false, state)?);
	ret.down = Some(Animation::new("data/succubus_down.png", false, state)?);
	ret.dancing = Some(Animation::new("data/succubus_walk_down.cfg", false, state)?);
	ret.up = Some(Animation::new("data/succubus_up.png", false, state)?);
	ret.left = Some(Animation::new("data/succubus_left.png", false, state)?);
	ret.right = Some(Animation::new("data/succubus_right.png", false, state)?);
	ret.kiss_down = Some(Animation::new("data/succubus_kiss_down.png", false, state)?);
	ret.kiss_up = Some(Animation::new("data/succubus_kiss_up.png", false, state)?);
	ret.kiss_left = Some(Animation::new("data/succubus_kiss_left.png", false, state)?);
	ret.kiss_right = Some(Animation::new("data/succubus_kiss_right.png", false, state)?);
	ret.walk_down = Some(Animation::new("data/succubus_walk_down.cfg", false, state)?);
	ret.walk_up = Some(Animation::new("data/succubus_walk_up.cfg", false, state)?);
	ret.walk_left = Some(Animation::new("data/succubus_walk_left.cfg", false, state)?);
	ret.walk_right = Some(Animation::new("data/succubus_walk_right.cfg", false, state)?);
	Ok(ret)
}

pub fn soldier_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/soldier_down.png", false, state)?);
	ret.down = Some(Animation::new("data/soldier_down.png", false, state)?);
	ret.up = Some(Animation::new("data/soldier_up.png", false, state)?);
	ret.left = Some(Animation::new("data/soldier_left.png", false, state)?);
	ret.right = Some(Animation::new("data/soldier_right.png", false, state)?);
	ret.dancing = Some(Animation::new("data/soldier_walk_down.cfg", false, state)?);
	ret.walk_down = Some(Animation::new("data/soldier_walk_down.cfg", false, state)?);
	ret.walk_up = Some(Animation::new("data/soldier_walk_up.cfg", false, state)?);
	ret.walk_left = Some(Animation::new("data/soldier_walk_left.cfg", false, state)?);
	ret.walk_right = Some(Animation::new("data/soldier_walk_right.cfg", false, state)?);
	Ok(ret)
}

pub fn devil_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/devil_down.png", false, state)?);
	ret.down = Some(Animation::new("data/devil_down.png", false, state)?);
	ret.up = Some(Animation::new("data/devil_up.png", false, state)?);
	ret.left = Some(Animation::new("data/devil_left.png", false, state)?);
	ret.right = Some(Animation::new("data/devil_right.png", false, state)?);
	ret.dancing = Some(Animation::new("data/devil_walk_down.cfg", false, state)?);
	ret.walk_down = Some(Animation::new("data/devil_walk_down.cfg", false, state)?);
	ret.walk_up = Some(Animation::new("data/devil_walk_up.cfg", false, state)?);
	ret.walk_left = Some(Animation::new("data/devil_walk_left.cfg", false, state)?);
	ret.walk_right = Some(Animation::new("data/devil_walk_right.cfg", false, state)?);
	Ok(ret)
}

pub fn bullet_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/cross.cfg", false, state)?);
	Ok(ret)
}

pub fn soldier_corpse_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/soldier_dead.png", false, state)?);
	Ok(ret)
}

pub fn devil_corpse_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/devil_dead.png", false, state)?);
	Ok(ret)
}

pub fn heart_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/heart.png", false, state)?);
	Ok(ret)
}

pub fn kiss_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/kiss.cfg", true, state)?);
	Ok(ret)
}

pub fn stun_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/stun.cfg", true, state)?);
	Ok(ret)
}

pub fn devil_hit_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/blood_splatter.cfg", true, state)?);
	Ok(ret)
}

pub fn bullet_hit_animations(state: &mut GameState) -> Res<Animations>
{
	let mut ret = Animations::new();
	ret.default = Some(Animation::new("data/cross_explosion.cfg", true, state)?);
	Ok(ret)
}

#[derive(Debug, Copy, Clone, PartialEq, Eq)]
pub enum CollisionClass
{
	None,
	Other,
	Wall,
	Projectile
}

impl CollisionClass
{
	pub fn collides_with(&self, other: CollisionClass) -> bool
	{
		match (*self, other)
		{
			(CollisionClass::Projectile, CollisionClass::Wall) => true,
			(CollisionClass::Other, CollisionClass::Wall) => true,
			(CollisionClass::Other, CollisionClass::Other) => true,
			_ => false,
		}
	}
}

object!
{
	Object
	{
		parent: usize = 0,

		pos: Option<Vec2D> = None,
		old_pos: Option<Vec2D> = None,

		vel:  Option<Vec2D> = None,
		animations: Option<Animations> = None,
		
		face_direction: Option<FaceDirection> = None,
		moving: bool = false,
		left: bool = false,
		right: bool = false,
		up: bool = false,
		down: bool = false,
		is_player: bool = false,
		collision: Option<Collision> = None,
		team: Option<Team> = None,
		attacking: bool = false,
		attack_time: f64 = 0.0,
		is_bullet: bool = false,
		directional_movement: bool = false,
		time_to_die: f64 = 0.0,
		health: Option<i32> = None,
		is_ai: bool = false,
		damage: i32 = 0,
		effect_size: f32 = 0.0,
		time_to_move: f64 = 0.0,
		collision_class: CollisionClass = CollisionClass::None,
		kissed: bool = false,
		is_kiss: bool = false,
		draw_layer: i32 = 0,
		speed: f32 = 0.0,
		kiss_until: f64 = 0.0,
		stunned_until: f64 = 0.0,
		is_damage_display: bool = false,
		num_hearts: i32 = 0,
		is_heart: bool = false,
		
		game_mode: Option<GameMode> = None,
		menu_mode: Option<MenuMode> = None,
		intro_mode: Option<IntroMode> = None,
	}
}

impl HasId for Object
{
	fn get_id(&self) -> usize
	{
		self.id.get()
	}
}

impl HasId for RefCell<Object>
{
	fn get_id(&self) -> usize
	{
		self.borrow().id.get()
	}
}


slr_def!
{
	pub struct Options
	{
		pub manual_vsync: bool = false,
		pub fullscreen: bool = false,
		pub music: bool = true,
		pub editor: bool = false,
		pub sound_volume: f32 = 1.0,
		pub music_volume: f32 = 1.0,
		pub soldier_health: i32 = 15,
		pub soldier_speed: f32 = 25.0,
		pub devil_health: i32 = 15,
		pub devil_speed: f32 = 20.0,
		pub succubus_speed: f32 = 75.0,
		pub succubus_hearts: i32 = 3,
		pub heart_heal: i32 = 5,
		pub heart_pickup: i32 = 3,
	}
}

pub struct GameState
{
	new_objects: Vec<RefCell<Object>>,
	ids_to_remove: Vec<usize>,
	
	pub id_mint: IdMint,
	
	pub core: Core,
	pub prim: PrimitivesAddon,
	pub font: FontAddon,
	pub audio: AudioAddon,
	pub bitmap_cache: BitmapCache,
	pub sample_cache: SampleCache,
	pub draw_interp: f32,
	pub key_down: Option<KeyCode>,
	pub key_up: Option<KeyCode>,
	pub time: f64,
	pub time_count: i64,
	pub ui_font: Font,
	pub paused: bool,
	pub start_fade: bool,
	pub fade_stop_time: f64,
	pub quit: bool,
	pub options: Options,
	pub sfx: Sfx,
}

impl GameState
{
	pub fn new(core: Core, prim: PrimitivesAddon, font: FontAddon, audio: AudioAddon, options: Options) -> Self
	{
		let ui_font = Font::load_bitmap_font(&font, "data/font.png").expect("Couldn't create font from 'data/font.png'");
		let sfx = Sfx::new(&audio, options.sound_volume, options.music_volume, options.music);
		
		Self
		{
			new_objects: vec![],
			ids_to_remove: vec![],
			id_mint: IdMint::new(),
			
			sfx: sfx,
			
			core: core,
			prim: prim,
			font: font,
			audio: audio,
			ui_font: ui_font,
			bitmap_cache: BitmapCache::new(),
			sample_cache: SampleCache::new(),
			draw_interp: 0.0,
			key_down: None,
			key_up: None,
			time: 0.0,
			time_count: 0,
			paused: false,
			start_fade: false,
			fade_stop_time: 0.0,
			quit: false,
			options: options,
		}
	}
		
	pub fn create_object(&mut self) -> Object
	{
		Object::new(self.id_mint.new_id())
	}

	pub fn add_object(&mut self, obj: Object)
	{
		self.new_objects.push(RefCell::new(obj));
	}

	pub fn remove_object(&mut self, id: usize)
	{
		for &old_id in &self.ids_to_remove
		{
			if old_id == id
			{
				return;
			}
		}
		self.ids_to_remove.push(id);
	}

	pub fn new_id(&mut self) -> UniqueId
	{
		self.id_mint.new_id()
	}
	
	pub fn play_sound(&mut self, sound: &str) -> Res<()>
	{
		self.sfx.play(&*self.sample_cache.load(&self.audio, sound)?, &self.audio);
		Ok(())
	}
}

impl WorldState<RefCell<Object>> for GameState
{
	fn get_new_objects(&mut self) -> &mut Vec<RefCell<Object>>
	{
		&mut self.new_objects
	}

	fn get_ids_to_remove(&mut self) -> &mut Vec<usize>
	{
		&mut self.ids_to_remove
	}
}
