// Copyright 2017 SiegeLord
// Licensed under GPL, see LICENSE for full terms

#![allow(unused_doc_comment)]

#[macro_use]
extern crate allegro;
extern crate cgmath;
extern crate allegro_dialog;
extern crate allegro_font;
extern crate allegro_image;
extern crate allegro_audio;
extern crate allegro_acodec;
extern crate allegro_primitives;
extern crate libc;
#[macro_use]
extern crate slr_config;
extern crate rand;
#[macro_use]
extern crate error_chain;
extern crate tiled;

use allegro::*;
use allegro_dialog::*;
use allegro_font::*;
use allegro_image::*;
use allegro_audio::*;
use allegro_acodec::*;
use allegro_primitives::*;
use util::load_config;
use game_state::{Object, GameState, Options};
use std::cmp::{min, max};
use engine::world::World;
use std::cell::RefCell;

mod errors
{
	error_chain!
	{
		types
		{
			 Error, ErrorKind, ResultExt, Res;
		}
		
		foreign_links
		{
			Tiled(::tiled::TiledError);
		}
	}
}

mod util;
#[macro_use]
mod game_state;
mod game_mode;
mod menu_mode;
mod engine;
mod animation;
mod actor;
mod ai;
mod parent;
mod sfx;
mod intro_mode;

static DT: f64 = 1.0 / 60.0;
static BUFFER_WIDTH: i32 = 640;
static BUFFER_HEIGHT: i32 = 480;

fn game()
{
	let mut options = Options::new();
	load_config("options.cfg", &mut options).expect("Could not load/parse 'options.cfg'");
	
	let core = Core::init().unwrap();
	let font = FontAddon::init(&core).expect("Could not init font addon");
	let _image = ImageAddon::init(&core).expect("Could not init image addon");
	let audio = AudioAddon::init(&core).expect("Could not init audio addon");
	let _acodec = AcodecAddon::init(&audio).expect("Could not init acodec addon");
	let prim = PrimitivesAddon::init(&core).expect("Could not init primitives addon");
	
	core.install_keyboard().unwrap();
	core.install_mouse().unwrap();
	
	if !options.manual_vsync
	{
		core.set_new_display_option(DisplayOption::Vsync, 1, DisplayOptionImportance::Suggest);
	}
	if options.fullscreen
	{
		core.set_new_display_flags(FULLSCREEN_WINDOW | RESIZABLE);
	}
	else
	{
		core.set_new_display_flags(RESIZABLE);
	}
	let disp = Display::new(&core, BUFFER_WIDTH * 2, BUFFER_HEIGHT * 2).unwrap();
	disp.set_window_title("Succubus Sans Frontieres");
	let bw = BUFFER_WIDTH;
	let bh = BUFFER_HEIGHT;
	let buffer = Bitmap::new(&core, bw, bh).unwrap();
	let cross_fade_buffer = Bitmap::new(&core, bw, bh).unwrap();
	core.clear_to_color(Color::from_rgb(0, 0, 0));
	core.flip_display();

	let timer = Timer::new(&core, DT).unwrap();

	let mut q = EventQueue::new(&core).unwrap();
	q.register_event_source(disp.get_event_source());
	q.register_event_source(core.get_keyboard_event_source().unwrap());
	q.register_event_source(core.get_mouse_event_source().unwrap());
	q.register_event_source(timer.get_event_source());
	
	let state = GameState::new(core, prim, font, audio, options);
	
	let mut world = World::<RefCell<Object>, GameState>::new(state);
	
	world.add_draw_behavior(Box::new(game_mode::GameModeDraw));
	world.add_draw_behavior(Box::new(menu_mode::MenuModeDraw));
	world.add_draw_behavior(Box::new(intro_mode::IntroModeDraw));
	world.add_draw_behavior(Box::new(actor::ActorDraw));
	world.add_draw_behavior(Box::new(actor::DamageDraw));
	world.add_draw_behavior(Box::new(game_mode::GameModeUIDraw));
	//~ world.add_draw_behavior(Box::new(actor::CollisionDraw));
	
	world.add_logic_behavior(Box::new(actor::BulletDeath));
	world.add_logic_behavior(Box::new(actor::HeartPickup));
	world.add_logic_behavior(Box::new(actor::TimeToDie));
	world.add_logic_behavior(Box::new(ai::ActorAI));
	world.add_logic_behavior(Box::new(actor::ActorDirectionalMovement));
	world.add_logic_behavior(Box::new(actor::ActorPhysics));
	world.add_logic_behavior(Box::new(actor::HealthDeath));
	world.add_logic_behavior(Box::new(actor::OldPos));
	world.add_logic_behavior(Box::new(game_mode::GameModeLogic));
	world.add_logic_behavior(Box::new(parent::ParentLogic));

	world.add_input_behavior(Box::new(actor::PlayerInput));
	world.add_input_behavior(Box::new(game_mode::GameModeInput));
	world.add_input_behavior(Box::new(menu_mode::MenuModeInput));
	world.add_input_behavior(Box::new(intro_mode::IntroModeInput));
	
	//~ let game_mode = create_game_mode("data/level0.cfg", &mut world.state).unwrap();
	//~ world.state.add_object(game_mode);
	let mode = menu_mode::create_menu_mode(&mut world.state).unwrap();
	world.state.add_object(mode);

	timer.start();
	let mut game_time = 0.0;
	let offset = world.state.core.get_time();
	'exit: loop
	{
		let scale = max(min(disp.get_width() / bw, disp.get_height() / bh), 1);
		let dx = ((disp.get_width() - bw * scale) / 2) as f32;
		let dy = ((disp.get_height() - bh * scale) / 2) as f32;

		for event in &mut q
		{
			world.state.key_down = None;
			world.state.key_up = None;
			
			match event
			{
				DisplayClose{..} =>
				{
					break 'exit;
				},
				DisplayResize{..} =>
				{
					disp.acknowledge_resize().unwrap();
				},
				KeyDown{keycode: k, ..} =>
				{
					world.state.key_down = Some(k);
					world.input().unwrap();
				},
				KeyUp{keycode: k, ..} =>
				{
					world.state.key_up = Some(k);
					world.input().unwrap();
				},
				TimerTick{count, ..} =>
				{
					if !world.state.paused
					{
						world.state.time_count += 1;
						world.state.time = world.state.time_count as f64 * DT;
						game_time = count as f64 * DT;
					}
					world.logic().unwrap();
				},
				_ => ()
			}
			
			if world.state.quit
			{
				break 'exit;
			}
		}
		
		if world.state.start_fade
		{
			world.state.start_fade = false;
			world.state.fade_stop_time = world.state.time + 1.0;
			world.state.core.set_target_bitmap(&cross_fade_buffer);
			world.state.core.draw_bitmap(&buffer, 0.0, 0.0, Flag::zero());
		}
		
		world.state.core.set_target_bitmap(&buffer);
		world.state.core.clear_to_color(Color::from_rgb(0, 0, 0));

		if !world.state.paused
		{
			let cur_time = world.state.core.get_time();
			world.state.draw_interp = ((cur_time - offset - game_time) / DT as f64) as f32;
		}
		world.draw().unwrap();
		
		world.state.core.set_target_bitmap(disp.get_backbuffer());
		world.state.core.clear_to_color(Color::from_rgb(0, 0, 0));
		
		world.state.core.draw_scaled_bitmap(&buffer, 0.0, 0.0, bw as f32, bh as f32, dx, dy, (bw * scale) as f32, (bh * scale) as f32, Flag::zero());
		
		if world.state.fade_stop_time > world.state.time
		{
			let a = ((world.state.fade_stop_time - world.state.time) / 1.0) as f32;
			let color = Color::from_rgba_f(a, a, a, a);
			world.state.core.draw_tinted_scaled_bitmap(&cross_fade_buffer, color, 0.0, 0.0, bw as f32, bh as f32, dx, dy, (bw * scale) as f32, (bh * scale) as f32, Flag::zero());
		}
		
		if world.state.options.manual_vsync
		{
			world.state.core.wait_for_vsync().unwrap();
		}
		world.state.core.flip_display();
	}
}

allegro_main!
{
	use std::panic::catch_unwind;

	match catch_unwind(game)
	{
		Err(e) =>
		{
			let err: String = e.downcast_ref::<&'static str>().map(|&e| { e.to_owned()}).or_else(||
			{
				e.downcast_ref::<String>().map(|e| e.clone())
			}).unwrap_or("Unknown error!".to_owned());

			println!("{}", err);
			show_native_message_box(None, "Error!", "An error has occurred! See game.log for more info.", &err, Some("You make me sad."), MESSAGEBOX_ERROR);
		}
		Ok(_) => ()
	}
}
